/* High-contrast overrides for Material for MkDocs */

/* Light mode — force black text on white background */
[data-md-color-scheme="default"] {
  --md-default-bg-color: #fff;
  --md-default-fg-color: #000;
  --md-default-fg-color--light: #222;
  --md-default-fg-color--lighter: #444;
  --md-typeset-color: #000;
  --md-code-fg-color: #000;
  --md-code-bg-color: #f0f0f0;
}

[data-md-color-scheme="default"] .md-typeset,
[data-md-color-scheme="default"] .md-typeset p,
[data-md-color-scheme="default"] .md-typeset li,
[data-md-color-scheme="default"] .md-typeset td,
[data-md-color-scheme="default"] .md-typeset th,
[data-md-color-scheme="default"] .md-typeset dt,
[data-md-color-scheme="default"] .md-typeset dd {
  color: #000 !important;
}

[data-md-color-scheme="default"] .md-typeset h1,
[data-md-color-scheme="default"] .md-typeset h2,
[data-md-color-scheme="default"] .md-typeset h3,
[data-md-color-scheme="default"] .md-typeset h4,
[data-md-color-scheme="default"] .md-typeset h5,
[data-md-color-scheme="default"] .md-typeset h6 {
  color: #000 !important;
}

[data-md-color-scheme="default"] .md-typeset code:not(.highlight code) {
  color: #000 !important;
}

[data-md-color-scheme="default"] .md-content {
  background-color: #fff;
}

[data-md-color-scheme="default"] .md-main {
  background-color: #fff;
}

/* Dark mode */
[data-md-color-scheme="slate"] {
  --md-default-fg-color: rgba(255, 255, 255, 0.9);
  --md-default-fg-color--light: rgba(255, 255, 255, 0.7);
  --md-default-fg-color--lighter: rgba(255, 255, 255, 0.5);
  --md-typeset-color: rgba(255, 255, 255, 0.9);
  --md-code-fg-color: #e0e0e0;
  --md-code-bg-color: #1a1a1a;
}
